'''
Copyright 2012 Research Institute eAustria

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

@author: Marian Neagul <marian@ieat.ro>
@contact: marian@ieat.ro
@copyright: 2012 Research Institute eAustria
'''

import logging

from me2.server.execution_manager import ExecutionManagerBase, ContainerSupervisor

from mjsrpc2 import RPCBase, RPCService, UnixSocketConnector, jsonmethod, jsonattribute
from me2.packaging.package import expand_variables

from mako.template import Template
from pkg_resources import ResourceManager

import os
from lockfile import LockFile
import threading
import Queue
import pipes
import subprocess
import time
import sys
import shutil
from threading import ThreadError
import uuid

resource_manager = ResourceManager()

class LXCContainer(RPCBase):
    lxc_config_tmpl = Template(text = resource_manager.resource_string(__name__, "data/templates/lxc_config.mak"))
    lxc_fstab_tmpl = Template(text = resource_manager.resource_string(__name__, "data/templates/lxc_fstab.mak"))
    isStarted = False
    isRunning = False
    isStopped = True

    def __init__(self, context, lxc_manager, basedir, configdir, ident, config):
        super(LXCContainer, self).__init__()
        self.context = context
        self.log = logging.getLogger("LXCContainer").getChild("ident-%s" % ident)
        self._lxc_manager = lxc_manager
        self.basedir = basedir
        self.configdir = configdir
        self.ident = ident
        self.runtime_config = config
        self.rootfs = os.path.join(basedir, "rootfs")
        self.tools_dir = self.rootfs + "/" + expand_variables("${execution.options.tools-dir}", self._lxc_manager.config)
        self.lxc_config_file = os.path.join(self.configdir, "lxc.config")
        self.lxc_fstab_file = os.path.join(self.configdir, "lxc.fstab")
        self.socket_dir = os.path.join(self.configdir, "net")
        self.socket_file = os.path.join(self.socket_dir, "me2-server.sock")
        self.lock = LockFile(os.path.join(self.basedir))
        self.lxc_start_container_command_template = expand_variables("${execution.start-container}", self._lxc_manager.config)
        self.lxc_stop_container_command_template = expand_variables("${execution.stop-container}", self._lxc_manager.config)
        self.lxc_info_container_command_template = expand_variables("${execution.container-info}", self._lxc_manager.config)
        self._rpc_server = None

    def start(self):
        self.log.info("Starting up container")
        self.prepare()
        self._start_rpc_server()
        self._start()

    def _start(self):
        command = self.lxc_start_container_command_template
        command = command % {'container_name': pipes.quote(self.ident),
                             'lxc_config': pipes.quote(self.lxc_config_file)
                             }
        command = self._lxc_manager._elevate(command)
        self.log.debug("Running command: %s", command)
        subprocess.check_call(command, shell = True)
        self.isStarted = True

    def _start_rpc_server(self):
        if self._rpc_server is not None:
            self.log.warn("RPC Server is already started")
            return
        self.log.info("Starting RPC Service. Socket: %s", self.socket_file)
        self._rpc_service = RPCService(self)
        self._rpc_server = UnixSocketConnector(rpc = self._rpc_service, socket_path = self.socket_file)

    def prepare(self):
        self.log.info("Preparing container")
        self.lxc_config()
        self.lxc_fstab()
        self._copy_agent_bundle()

    def _copy_agent_bundle(self):
        try:
            os.makedirs(self.tools_dir)
        except OSError, e:
            pass
        dest_dir = os.path.join(self.tools_dir, "agent.pybundle")
        self.log.info("Copying agent to %s", dest_dir)
        shutil.copyfile(self.context.agent_bundle, dest_dir)

    def lxc_config(self):
        config = {'hostname': self.ident,
                  'rootfs': self.rootfs,
                  'fstab': self.lxc_fstab_file,
                  'agent_bundle': self.context.agent_bundle
                  }
        content = self.lxc_config_tmpl.render(**config)
        self.log.info("Saving LXC Config as: %s", self.lxc_config_file)
        with open(self.lxc_config_file, "w") as f:
            f.write(content)

        if not os.path.exists(self.socket_dir):
            os.makedirs(self.socket_dir, 0755)

    def lxc_fstab(self):
        config = {'rootfs': self.rootfs,
                  'host_mde_socketdir': self.socket_dir,
                  'agent_bundle': self.context.agent_bundle
                  }

        content = self.lxc_fstab_tmpl.render(**config)
        self.log.debug("Saving fstab to %s", self.lxc_fstab_file)
        with open(self.lxc_fstab_file, "w") as f:
            f.write(content)


class LxcExecutionManager(ExecutionManagerBase):
    log = logging.getLogger("LxcExecutionManager")

    def __init__(self, *args, **kw):
        super(LxcExecutionManager, self).__init__(*args, **kw)
        self.registry = ContainerSupervisor() # This should go to execution manager base
        self.lxc_info_container_command_template = expand_variables("${execution.container-info}", self.config)

    @jsonmethod
    @jsonattribute(name = "bottle_dir", kind = str, documentation = "The bottle directory in which the container should be started")
    @jsonattribute(name = "config_dir", kind = str, documentation = "The config directory for the bottle")
    @jsonattribute(name = "ident", kind = str, documentation = "The identifier of the bottle")
    @jsonattribute(name = "config", kind = object, documentation = "The running config")
    def start(self, bottle_dir, config_dir, ident, config):
        with self.lock:
            container = LXCContainer(self.context, self, bottle_dir, config_dir, ident, config)
            container = self.start_container(container)
            self.registry.add(container)
        return container

    @jsonmethod
    @jsonattribute(name = "container_id", kind = str, documentation = "The is of the container that should be stoped")
    def stop(self, container_id):
        with self.lock:
            container = self.registry.get(container_id)
            container.stop()

    @jsonmethod
    @jsonattribute(name = "container_id", kind = str, documentation = "The id of the container")
    def info(self, container_id):
        """Method for checking the status of a container. 
        @return: an dict with the information
        """
        container_id = uuid.UUID(container_id).get_hex()
        with self.lock:
            lxc_info = self.get_lxc_info(container_id)
            managed = False
            try:
                container = self.registry.get(container_id)
                if container:
                    managed = True
            except KeyError:
                pass
            lxc_info['managed'] = managed
            return lxc_info

    @jsonmethod
    def list(self):
        pass


    def start_container(self, container):
        self.log.debug("Starting container")
        container.start()
        return container

    def stop_container(self, container):
        self.log.debug("Stoping container")
        container.stop()
        return container

    def get_lxc_info(self, container_id):
        log = self.log.getChild("get_lxc_info")
        command = self._elevate(self.lxc_info_container_command_template % {'container_name':container_id})
        output = subprocess.check_output(command, shell = True)
        output = output.split("\n")
        output = [line.strip().split(":") for line in output if len(line) > 0]
        output = [line for line in output if len(line) == 2]
        output = [ (name.strip(), value.strip()) for name, value in output]
        output = dict(output)
        log.debug("Got output from lxc list: %s", output)
        running = output["state"] == "RUNNING"

        result = {'running': running}
        if running and 'pid' in output:
            result['pid'] = int(output['pid'])
        return result



'''
Copyright 2012 Research Institute eAustria

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

@author: Marian Neagul <marian@ieat.ro>
@contact: marian@ieat.ro
@copyright: 2012 Research Institute eAustria
'''

import os
import uuid
import logging
import shutil
import pipes
import subprocess
import grp
import pwd

from mjsrpc2 import RPCBase, jsonmethod, jsonattribute

from me2.server.util import OperationResult, ContainerManagerException
from me2.packaging.package import expand_variables

class ContainerBottle(object):
    uuid = None
    path = None
    rootfs = None
    configdir = None

class StorageManagerException(ContainerManagerException, IOError):
    pass

class StorageManagerBase(RPCBase):
    log = logging.getLogger("StorageManagerBase")
    def __init__(self, context, manager, config):
        RPCBase.__init__(self)
        self.config = config
        self._manager = manager
        self._elevate_command = expand_variables("${elevate-privileges}", self.config, default = "")
        self.root_directory = expand_variables("${storage.root-directory}", config)
        if not os.path.exists(self.root_directory): self.create_subvolume(self.root_directory)
        base_directory = config.get("base-directory", "base")
        self.base_directory = os.path.join(self.root_directory, base_directory)
        if not os.path.exists(self.base_directory): self.create_subvolume(self.base_directory)
        instances_directory = config.get("instances-directory", 'instances')
        self.instances_directory = os.path.join(self.root_directory, instances_directory)
        if not os.path.exists(self.instances_directory): self.create_subvolume(self.instances_directory)
        self.log.debug("Root-directory: %s", self.root_directory)
        self.log.debug("Base-directory: %s", self.base_directory)
        self.log.debug("Instances directory: %s", self.instances_directory)

    def isbottle(self, directory):
        log = self.log.getChild("isbottle")
        if not os.path.exists(directory):
            log.info("Directory %s does not exist", directory)
            return False
        config_dir = os.path.join(directory, "config")
        if not os.path.exists(config_dir):
            log.info("Config directory %s is missing", config_dir)
            return False
        rootfs_dir = os.path.join(directory, "rootfs")
        if not os.path.exists(rootfs_dir):
            log.info("Rootfs directory %s is missing")
            return False
        try:
            uuid.UUID(os.path.basename(directory))
        except ValueError:
            log.exception("Invalid identifier for directory")
            return False

        parent_link = os.path.join(directory, "base")
        if not os.path.exists(parent_link): return False
        return True

    @jsonmethod
    @jsonattribute(name = "name", kind = str)
    def create_base_bottle(self, name):
        self.log.debug("Creating bottle %s", name)
        bottle_location = self._build_bottle_path(name)
        self.log.info("Creating new bottle in %s", bottle_location)
        if self.exists(name = name):
            self.log.warn("Bottle %s already exists!", name)
            return bottle_location
        try:
            self.create_subvolume(bottle_location)
        except:
            self.log.exception("Failed to create bottle %s cleaning up", name)
            self.delete_subvolume(bottle_location)
            raise # Rethrow the exception
        return bottle_location


    def promote_bottle(self, new_name, source):
        pass

    def _build_bottle_path(self, name):
        return os.path.join(self.base_directory, pipes.quote(name))

    @jsonmethod
    @jsonattribute(name = "name")
    def exists(self, name):
        path = self._build_bottle_path(name)
        if os.path.exists(path):
            return True
        else: # ToDo: really check that the file does not exist as os.path.exists return false if it can stat the path
            return False

    @jsonmethod
    def list_base_bottles(self):
        result = OperationResult()
        bottles = []
        status = False
        try:
            bottles = os.listdir(self.base_directory)
        finally:
            result.status = status
            result.result = bottles
        return vars(result)

    @jsonmethod
    def list_bottles(self):
        log = self.log.getChild("list_bottles")
        manager = self._manager
        lxc_manager = manager.lxc
        print dir(lxc_manager)
        print self.config
        response = OperationResult()
        bottle_names = [ subdir for subdir in os.listdir(self.instances_directory) if self.isbottle(os.path.join(self.instances_directory, subdir))]
        bottles = []
        for name in bottle_names[:]:
            bottle = {}
            bottle["uuid"] = str(uuid.UUID(name))
            base = os.path.basename(os.path.realpath(os.path.join(self.instances_directory, name, "base")))
            bottle["base-bottle"] = base
            bottles.append(bottle)
        response.status = True
        response.result = bottles
        return vars(response)

    @jsonmethod
    @jsonattribute(name = "base")
    def clone(self, base, bottle_id = None):
        bottle = ContainerBottle()
        if bottle_id is None:
            bottle_id = uuid.uuid4().get_hex()

        bottle.uuid = bottle_id
        clone_path = os.path.join(self.instances_directory, bottle_id)
        if os.path.exists(clone_path):
            raise StorageManagerException("Bottle already exists")
        bottle.path = clone_path
        bottle.rootfs = os.path.join(bottle.path, "rootfs")
        bottle.configdir = os.path.join(bottle.path, "config")

        base_container_path = os.path.join(self.base_directory, base)
        self.log.debug("Creating a new bottle in %s", bottle.path)
        self.create_snapshot(base_container_path, bottle.path)
        os.symlink(base_container_path, os.path.join(bottle.path, "base"))
        os.makedirs(bottle.configdir, mode = 0755)
        self.log.debug("Bottle dict: %s", vars(bottle))
        return vars(bottle)

    @jsonmethod
    @jsonattribute(name = "bottle_id", kind = str)
    def remove(self, bottle_id):
        hex_id = uuid.UUID(bottle_id).get_hex()
        bottle_path = os.path.join(self.instances_directory, hex_id)
        try:
            assert self.isbottle(bottle_path)
            self.delete_snapshot(bottle_path)
            if os.path.exists(bottle_path):
                shutil.rmtree(bottle_path)
        except:
            self.log.exception("Exception trying to remove bottle %s" % bottle_id)
            return False
        return True

    @jsonmethod
    @jsonattribute(name = "name")
    def remove_base_bottle(self, name):
        path = self._build_bottle_path(name)
        self.delete_subvolume(path)

    def create_snapshot(self, source, destination):
        """Creates a snapshot of source in destination
        
        @param source: The source that should be snapshot
        @param destination: The directory where the snapshot should go
        """
        raise NotImplementedError()

    def delete_snapshot(self, path):
        """Deletes a snapshot
        
        @param path: The path to the snapshot
        """
        raise NotImplementedError()

    def _elevate(self, command):
        return self._elevate_command % {'command': command}


    def create_subvolume(self, path):
        raise NotImplementedError()

    def delete_subvolume(self, path):
        raise NotImplementedError()

class BtrfsStorageManager(StorageManagerBase):
    log = logging.getLogger("BtrfsStorageManager")
    def __init__(self, *args, **kw):
        StorageManagerBase.__init__(self, *args, **kw)

    def set_permission(self, path, perms):
        if not os.path.abspath(path).startswith(self.root_directory):
            raise StorageManagerException("Path (%s) is not within me2 root directory", self.root_directory)
        command_template = expand_variables("${storage.chmod-command}", self.config)
        command = command_template % {"permissions": perms,
                                      "path": pipes.quote(path)}
        command = self._elevate(command)
        subprocess.check_call(command, shell = True)

    def set_owner(self, path, user):
        uid = pwd.getpwnam(user)[2]
        self._set_owner_uid(path, uid)
        raise NotImplementedError()

    def _set_owner_uid(self, path, uid):
        command_template = None
        raise NotImplementedError()

    def set_group(self, path, group):
        gid = grp.getgrnam(group)
        self._set_group_uid(path, gid)
        raise NotImplementedError()

    def _set_group_uid(self, path, gid):
        command = None
        raise NotImplementedError()

    def create_subvolume(self, path):
        path = os.path.abspath(path)
        if not path.startswith(self.root_directory):
            raise StorageManagerException("Path (%s) is not within me2 root directory", self.root_directory)
        if os.path.lexists(path):
            raise StorageManagerException("Subvolume already exists")

        self.log.debug("Creating a new subvolume in %s", path)
        volume_name = os.path.basename(path)
        directory_name = os.path.dirname(path)
        command_template = expand_variables("${storage.create-volume}", self.config)
        command = command_template % {'destination_path': pipes.quote(directory_name),
                                      'subvolume_name': pipes.quote(volume_name)}
        command = self._elevate(command)
        subprocess.check_call(command, shell = True)
        self.set_permission(path, expand_variables("${storage.volume-permissions}", self.config))

    def delete_subvolume(self, path):
        self.log.debug("Delete-ing subvolume %s", path)
        path = os.path.abspath(path)
        if not path.startswith(self.root_directory):
            raise StorageManagerException("Path is not within root directory")
        command_template = expand_variables("${storage.delete-volume}", self.config)
        command = command_template % {"subvolume_path": path}
        command = self._elevate(command)
        subprocess.check_call(command, shell = True)

    def create_snapshot(self, source, destination):
        self.log.debug("Creating BTRFS snapshot from %s to %s", source, destination)
        command = self.config.get('storage', {}).get("create-snapshot") % {'source_path': pipes.quote(source),
                                                                           'destination_path': pipes.quote(destination)}
        elevate_privileges = self.config.get('elevate-privileges', None)
        if elevate_privileges:
            command = elevate_privileges % {"command": command}
        self.log.debug("Command to be executed: %s", command)
        subprocess.check_call(command, shell = True)
        self.set_permission(pipes.quote(destination), expand_variables("${storage.volume-permissions}", self.config))

    def delete_snapshot(self, path):
        #ToDo: I should check that the absolute path is inside the base directory
        self.log.debug("Deleting BTRFS snapshot from %s", path)
        command = self.config.get('storage', {}).get('delete-snapshot') % {'path': pipes.quote(path)}
        elevate_privileges = self.config.get('elevate-privileges', None)
        if elevate_privileges:
            command = elevate_privileges % {"command": command}
        self.log.debug("Command to be executed: %s", command)
        subprocess.check_call(command, shell = True)



'''
Copyright 2012 Research Institute eAustria

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

@author: Marian Neagul <marian@ieat.ro>
@contact: marian@ieat.ro
@copyright: 2012 Research Institute eAustria
'''

from Queue import Queue


class QueuePair(object):
    def __init__(self, in_q = None, out_q = None):
        if in_q is None:
            self.inq = Queue()
        else:
            self.inq = in_q
        if out_q is None:
            self.outq = Queue()
        else:
            self.outq = out_q

    def put(self, item, block = True, timeout = None):
        self.outq.put(item, block, timeout)

    def get(self, block = True, timeout = None):
        self.inq.get(block, timeout)

    def invert(self):
        return QueuePair(in_q = self.outq, out_q = self.outq)



'''
Copyright 2012 Research Institute eAustria

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

@author: Marian Neagul <marian@ieat.ro>
@contact: marian@ieat.ro
@copyright: 2012 Research Institute eAustria
'''


import logging

from me2.packaging.dric import MetaRepository
from me2.packaging.package import parse_artifact_spec
from me2.packaging.package import Bundle

from mjsrpc2 import RPCBase, jsonmethod, jsonattribute


class BundleManager(RPCBase):
    log = logging.getLogger("BundleManager")
    def __init__(self, context, manager, config, storage_manager):
        RPCBase.__init__(self)
        self._manager = manager
        self.context = context
        self._config = config
        self._storage_manager = storage_manager
        self.repo = MetaRepository(repositorys = self._config.get("sources", []))

    @jsonmethod
    @jsonattribute(name = "bundle_specification", type = str)
    def import_bundle(self, bundle_specification):
        self._import_bundle(bundle_specification)

    def _import_bundle(self, bundle_specification):
        groupId, artifactId, version, classifier = parse_artifact_spec(bundle_specification)
        # Download file
        local_file = self.repo.fetch(groupId, artifactId, version, classifier)
        bundle = Bundle(local_file)
        # Create the base bundle
        bottle_location = self._storage_manager.create_base_bottle(name = bundle_specification)
        # Extract the content
        self.explode_bundle(bundle, bottle_location)
        return bottle_location



    def explode_bundle(self, bundle, destination):
        self.log.info("Exploding %s to %s", bundle, destination)
        bundle.extract(destination)





'''
Copyright 2012 Research Institute eAustria

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

@author: Marian Neagul <marian@ieat.ro>
@contact: marian@ieat.ro
@copyright: 2012 Research Institute eAustria
'''

import threading
import logging
import sys
import os

from mjsrpc2 import RPCService, UnixSocketConnector

from me2.server.container import Deployer
from me2.common.util import QueuePair


class DaemonManager(threading.Thread):
    log = logging.getLogger("DaemonManager")
    def __init__(self, context):
        threading.Thread.__init__(self)
        self.context = context
        self.command_queue = QueuePair()
        self.deployer = Deployer(context = context, config_file = context.config_file)
        self.deployer_rpc = RPCService(self.deployer)

    def run(self):
        # Wait for incomming messages on command queue
        # The messages should be of the following form
        # {'command': '', **}
        self.log.info("Waiting for commands")
        context = self.context
        apps = context.application[:]

        self.start_connectors()
        self.start_apps(apps)

        while True:
            msg = self.command_queue.get(block = True)
            self.log.debug("Received command: %s", msg)
            if msg['command'] == 'shutdown':
                self.shutdown()
                break

    def start_apps(self, apps):
        pass

    def start_connectors(self):
        context = self.context
        if context.enable_unix_server:
            socketPath = context.unix_socket
            self.log.info("Starting UNIX Domain Socket Server on %s", socketPath)
            if os.path.exists(socketPath):
                self.log.error("Socket %s exists. It will be removed if possible", socketPath)
                os.remove(socketPath)
                self.log.debug("Socket %s was removed", socketPath)
            unix_connector = UnixSocketConnector(self.deployer_rpc, socketPath)
            unix_connector.setDaemon(True)
            unix_connector.start()
        if context.enable_tcp_server:
            self.log.info("Starting HTTP Server on %s:%s", self.context.tcp_bind, self.context.tcp_port)
            from me2.server.web import server
            from jsonrpc2 import JsonRpcApplication
#            app = JsonRpcApplication()
#            app.rpc = self.deployer_rpc
            app = JsonRpcApplication()
            app.rpc = self.deployer_rpc
            server.start_server(self.context.tcp_bind, self.context.tcp_port, self.deployer, app)
            self.log.info("HTTP Monster started!")

    def shutdown(self):
        pass


def main(context):
    manager = DaemonManager(context)
    logging.debug("Starting daemon manager")
    manager.setDaemon(True)
    manager.start()
    while True:
        try:
            if not manager.is_alive():
                break
            manager.join(timeout = 0.10)
        except KeyboardInterrupt:
            logging.critical("KeyboardInterrupt cought. Exiting!")
            sys.exit(1)
        except Exception:
            logging.exception("Cought exception")

